import psychopy.core
import psychopy.visual
import psychopy.logging
import psychopy.event
import numpy
from psychopy.hardware.emulator import SyncGenerator
import collections

class Clock(object):
    '''
    Time-keeping functionality for expcontrol by wrapping Psychopy's
    core.Clock instance.'''

    def __init__(self):
        '''Initialise a clock instance.'''
        self.ppclock = psychopy.core.Clock()
        super(Clock,self).__init__()
        psychopy.logging.setDefaultClock(self.ppclock)
        return

    def __call__(self):
        '''Return the current time stamp from ppclock.getTime'''
        return self.ppclock.getTime()

    def start(self):
        '''Reset the clock to 0.'''
        self.ppclock.reset()
        return self()

    def wait(self,time):
        psychopy.core.wait(time)
        return

    def waituntil(self,time):
        self.wait(time-self())
        return

class PulseClock(Clock):
    '''
    Time-keeping with tracking of pulses (e.g. from a scanner trigger)
    through a keyboard button at some interval. Note that time is
    still tracked in seconds, not pulses. So on its own, using this class
    will ensure that you synchronise your experiment to the first pulse
    (see start method), but everything afterwards still runs in seconds as
    with the standard Clock class.

    The only further refinement is that the clock will attempt to meausure
    pulse period empirically whenever given a chance (ie, self.waituntil is
    called with enough remaining time that a pulse is expected during the
    wait. These estimates are stored in self.periodhistory.
    '''
    def __init__(self,key,period,pulsedur=0.01,tolerance=.1,timeout=20., \
            winhand=None,verbose=False):
        self.period = period
        self.pulsedur = pulsedur
        self.tolerance = tolerance
        self.periodhistory = [period]
        self.timeout = timeout
        self.winhand = winhand
        self.verbose = verbose
        super(PulseClock,self).__init__()
        self.keyhand = KeyboardResponse(key,self.ppclock)
        return

    def waitpulse(self):
        calltime = self()
        k,t = self.keyhand.waitkey(self.timeout)
        assert k,'exceeded %.0fs timeout without receiving pulse' % \
                    self.timeout
        # first time of response if we got multiple
        keytime = t[0]
        return keytime

    def start(self):
        # flip and get keys to clear out any existing key presses
        self.winhand()
        k,t = self.keyhand()
        # need to first reset the second clock to make the timeout counter
        # in waitpulse work properly
        super(PulseClock,self).start()
        if self.verbose:
            print 'waiting for pulse...'
        # but this means that the starttime recorded here is off
        starttime = self.waitpulse()
        if self.verbose:
            print 'got it!'
        # so we adjust the clock to compensate for starttime (not quite the
        # same as zeroing the clock - if time has passed since the pulse
        # was received this operation will produce a current clock time >0
        self.ppclock.add(starttime)
        # return current time after all this
        return self()

    def waituntil(self,time):
        # current time
        now = self()
        nowpulse = now / self.period
        timepulse = time / self.period
        npulseleft = numpy.floor(timepulse)-numpy.floor(nowpulse)
        if npulseleft < 1:
            # less than a self.period left, so wait it out using standard
            # second clock
            super(PulseClock,self).waituntil(time)
            return
        # if we make it here, there must be pulses to catch
        actualtime = self.waitpulse()
        # we expect the next pulse to be number
        predictpulse = numpy.ceil(now / self.period)
        # and if that's true this is when it should happen
        predicttime = predictpulse * self.period
        # now we can update our estimate of period like so...
        newpulse = actualtime / predictpulse
        if numpy.abs(newpulse-self.period) > self.tolerance:
            raise Exception('pulse period beyond tolerance: ' +
                    'expected=%.4f, estimated=%.4f' % (self.period,newpulse))
        self.period = newpulse
        if self.verbose:
            print 'Pulse at %.2f. tr=%.3f' % (actualtime,newpulse)
        self.periodhistory.append(newpulse)
        # avoid catching the same pulse twice
        if (time-self()) > self.pulsedur:
            core.wait(self.pulsedur)
        # we recurse with a depth of npulseleft. This is important to
        # handle cases where you are waiting n pulses + a bit extra
        self.waituntil(time)
        return

class Window(object):
    '''
    Display control functionality for expcontrol by wrapping
    Psychopy's visual.Window.
    '''

    def __init__(self,*args,**kwargs):
        '''
        Initialise a window instance. All input arguments are piped to
        psychopy.visual.Window.
        '''
        self.winhand = psychopy.visual.Window(*args,**kwargs)
        # flip a few times because it is thought this helps stabilise
        # timings
        for i in xrange(50):
            self()
        return

    def __call__(self):
        return self.winhand.flip()

    def close(self):
        self.winhand.close()
        return

class KeyboardResponse(object):
    '''
    Psychopy-based keyboard response checking.
    '''
    esckey = 'escape'

    def __init__(self,keylist,clock):
        if not isinstance(keylist,collections.Iterable):
            keylist = [keylist]
        self.keylist = keylist + [self.esckey]
        self.ppclock = clock
        return

    def __call__(self):
        # map list of (k,t) tuples to one k tuples and one t tuple
        ktup = psychopy.event.getKeys(keyList=self.keylist,
            timeStamped=self.ppclock)
        return self.parsekey(ktup)

    def waitkey(self,dur=float('inf')):
        ktup = psychopy.event.waitKeys(maxWait=dur,keyList=self.keylist,
                timeStamped=self.ppclock)
        return self.parsekey(ktup)

    def parsekey(self,ktup):
        k = []
        t = []
        if ktup:
            k,t = zip(*ktup)
        if self.esckey in k:
            raise Exception('user pressed escape')
        return numpy.array(k),numpy.array(t)

class PulseEmulator(object):
    '''
    Simulate pulses at some period. Just a convenience wrapper for
    psychopy.hardware.emulator.SynchGenerator.
    '''
    def __init__(self,*args,**kwargs):
        self.pulsehand = SyncGenerator(*args,**kwargs)
        return

    def start(self):
        self.pulsehand.start()
        psychopy.core.runningThreads.append(self.pulsehand)
        return

    def stop(self):
        self.pulsehand.stop()
        return



# nb pylink must be obtained from the SR research forums and be on your
# pythonpath.
import pylink
import time
import numpy

class EyeLinkTracker(object):
    '''
    Handle common eye tracker tasks with a somewhat more intuitive
    interface than stock pylink.
    '''
    def __init__(self,size=[1024,768],calibscale=1.,ip='100.1.1.1', \
            bgcolor=[127,127,127],fgcolor=[255,255,255],targetdiameter=20, \
            targethole=5,calibrationtype='HV9',calibrationpacing=.9, \
            viewdistance=None,screenwidth=None):

        self.size = tuple(size)
        # connect to tracker and do initial config
        self.tracker = pylink.EyeLink(ip)
        # flush out any pending key presses and get back to offline mode in
        # case we crashed out while recording
        pylink.flushGetkeyQueue()
        self.tracker.setOfflineMode()
        self.fgcolor = fgcolor
        self.bgcolor = bgcolor
        self.targetdiameter = targetdiameter
        self.targethole = targethole

        # month, day, hour, minute.
        self.remotefilename = time.strftime('%m%d%H%M')
        self.tracker.openDataFile(self.remotefilename)

        self.calibsize = (numpy.array(self.size)*calibscale)
        calibarea = numpy.round(numpy.array(self.size)-self.calibsize)
        alldims = (calibarea[0],calibarea[1],self.calibsize[0],
                self.calibsize[1])
        self.tracker.sendCommand(
                'screen_pixel_coords =  %d %d %d %d' % alldims)
        self.tracker.sendMessage(
                "DISPLAY_COORDS  %d %d %d %d" % alldims)
        self.tracker.sendMessage(
                "SCREEN_COORDS  0 0 %d %d" % self.size)

        # for robustness we set a bunch of other parameters so that any
        # weird defaults get overwritten
        if viewdistance:
            self.tracker.sendCommand('simulation_screen_distance=%d' % \
                    (viewdistance*10))
            self.tracker.sendMessage('VIEW_DISTANCE %d' % \
                    (viewdistance*10))
        self.tracker.sendCommand('automatic_calibration_pacing=%d' % \
                (calibrationpacing*1000))
        if screenwidth:
            self.tracker.sendMessage('SCREEN_WIDTH %d' % screenwidth)
        # NB this command is necessary whenever changing
        # screen_pixel_coords
        self.tracker.sendCommand('calibration_type=' + calibrationtype)
        if (self.tracker.getTrackerVersion()== 2):
            self.tracker.sendCommand("select_parser_configuration 0")
        else:
            self.tracker.sendCommand("saccade_velocity_threshold = 35")
            self.tracker.sendCommand("saccade_acceleration_threshold = 9500")
            self.tracker.setFileEventFilter("LEFT,RIGHT,FIXATION,SACCADE,BLINK,MESSAGE,BUTTON")
            self.tracker.setFileSampleFilter("LEFT,RIGHT,GAZE,AREA,GAZERES,STATUS")
            self.tracker.setLinkEventFilter("LEFT,RIGHT,FIXATION,SACCADE,BLINK,BUTTON")
            self.tracker.setLinkSampleFilter("LEFT,RIGHT,GAZE,GAZERES,AREA,STATUS")
            self.tracker.sendCommand("button_function 5 'accept_target_fixation'")
        return

    def calibrate(self):
        # start the main calibration/validation interface
        pylink.openGraphics(self.size)
        # these commands cause a hard crash if sent before openGraphics
        pylink.setCalibrationColors(self.fgcolor,self.bgcolor)
        pylink.setTargetSize(self.targetdiameter,self.targethole)
        self.tracker.doTrackerSetup()
        self.eyeused = self.tracker.eyeAvailable()
        pylink.closeGraphics()
        return

    def start(self):
        err = self.tracker.startRecording(1,1,1,1)
        assert not err, 'EyeLink error: ' + err
        return

    def message(self,msg):
        self.tracker.sendMessage(msg)
        return

    def stop(self,outfile):
        # pumpDelay is a lower priority delay which does not block
        # background events. msecDelay is more aggressive. Here used to
        # catch last bit of data before stopping the recording
        pylink.pumpDelay(100)
        # idle mode
        self.tracker.setOfflineMode()
        pylink.msecDelay(500)
        # close the file on the tracker HD. Can take a while...
        self.tracker.closeDataFile()
        if outfile is not None:
            self.tracker.receiveDataFile(self.remotefilename,outfile)
        self.tracker.close()
        return



import numpy 
import datetime
from . import event
import functools

def addcustomdict(f):
    '''
    Decorator for save functionality (currently in Experiment class).
    '''
    @functools.wraps(f)
    def wrapper(*args,**kwargs):
        # a few assumptions here: 1) customdict is in kwargs, 2) res is in
        # args
        if 'customdict' in kwargs and kwargs['customdict']:
            # assigning to args requires list type
            args = list(args)
            args[1] = args[1].copy()
            for k,v in kwargs['customdict'].iteritems():
                args[1][k] = v
        return f(*args,**kwargs)
    return wrapper

class Controller(object):
    '''
    Control experiment timing, stimulus delivery and response collection.
    '''

    def __init__(self,window=None,response=None,clock=None,eyetracker=None):
        '''
        Initialise a controller instance. For example inputs, see
        expcontrol.psychopydep.window, KeyboardResponse and clock.'''
        self.window = window
        self.response = response
        self.clock = clock
        self.eyetracker = eyetracker
        return

    def __call__(self):
        '''
        Check for responses and flip the screen. If this is called often
        enough you will achieve sync with the screen refresh (assuming that
        your window method holds until the refresh).'''
        frametime = self.window()
        response,resptime = self.response()
        return response,resptime,frametime

class Experiment(object):
    '''
    Class for running a set of trials in some experiment.'''

    def __init__(self,conditions={},preevent=None,postevent=None,
            subject=None,context=None,verbose=False):
        '''
        Initialise an Experiment instance.

        Keyword arguments:
        conditions -- dict or list of Event-derived instances (including
            EventSeq).
        preevent -- An event-derived instance that is called before the
            main trial sequence with an endtime of numpy.inf. This instance
            should practically always be initialised with
            skiponresponse=True.
        postevent -- An event-derived instance that is called after the
            main trial sequence. Otherwise similar to preevent above.
        subject -- str for log file. Prompted if undefined.
        context -- str for log file. Prompted if undefined.
        verbose -- print various info.
        '''
        self.conditions = conditions
        self.preevent = preevent
        self.postevent = postevent
        if not subject:
            subject = raw_input('subject: ')
        self.subject = subject
        if not context:
            context = raw_input('context: ')
        self.context = context
        self.session = numpy.datetime64(datetime.datetime.now())
        return

    def __call__(self,controller,conditionkeys,seqclass=event.EventSeqAbsTime):
        '''
        Run a sequence of trials of the experiment, and return panda
        dataframes corresponding to the main trial sequence and the output
        of any pre/post events.

        Arguments:
        controller -- a Controller instance
        conditionkeys -- a list of keys or indices into self.conditions,
            which defines the sequence of conditions over the run.
        seqclass -- class to use for creating the trial sequence. Use
            EventSeqRelTime if absolute timing is not possible (e.g.,
            self-timed events, synching to pulses).
        '''
        # unpack to a fixed sequence of conditions
        # note that we leave name blank so we don't risk overwriting the
        # condition names in nested EventSeq-derived instances
        sequence = seqclass(
                [self.conditions[key] for key in conditionkeys],
                name=None)
        # run preevent, zero the clock
        preevlog = None
        if self.preevent:
            preevlog,preresplog = self.preevent(controller,numpy.inf)
            preevlog['subject'] = self.subject
            preevlog['session'] = self.session
            preevlog['context'] = self.context
        controller.clock.start()
        # main sequence
        eventlog,resplog = sequence(controller)
        # possible post-flight
        postevlog = None
        if self.postevent:
            postevlog,postresplog = self.postevent(controller,numpy.inf)
            postevlog['subject'] = self.subject
            postevlog['session'] = self.session
            postevlog['context'] = self.context
        eventlog['subject'] = self.subject
        eventlog['session'] = self.session
        eventlog['context'] = self.context
        return eventlog,resplog,preevlog,preresplog,postevlog,postresplog

    @addcustomdict
    def to_sql(self,res,path,customdict=None):
        import sqlalchemy
        self.engine = sqlalchemy.create_engine('sqlite:///' + path)
        res.to_sql(self.context,engine,if_exists='append')
        return

    @addcustomdict
    def to_hdf(self,res,path,customdict=None):
        res.to_hdf(path,self.context,append=True)
        return



import numpy
import pandas
import collections

eventkeys = ['name','condition','oncall','onframe','onend']
def prepeventrow(ind=None):
    '''
    return row(s) of the event log with columns as in eventkeys and indices
    as in ind.
    '''
    return pandas.DataFrame(columns=eventkeys,index=ind,dtype=float)

respkeys = ['key','onresponse_score','onresponse_rt']
def prepresprow(ind=None):
    '''
    return row(s) of the response log with columns as in respkey and indices
    as in ind.
    '''
    return pandas.DataFrame(columns=respkeys,index=ind,dtype=float)

class Event(object):
    '''
    The smallest independent element of an experiment. This class stubs out
    callbacks for different stages of the Event (oncall, onframe,
    onresponse, onend). The purpose of this is to support sub-classing
    where only some of these methods may be required.'''

    def __init__(self,name=None,duration=0,skiponresponse=[],verbose=False):
        '''
        Initialise an Event instance.
        
        Keyword arguments:
        name --- is placed in the name field of the return from the call
            method.
        duration --- length of event in controller.clock units. Note that
            the duration field is not used directly here in this class, but
            is referenced elsewhere to set the endtime argument correctly
            at calltime.  
        skiponresponse -- if skiponresponse, the Event is terminated early
            when a response is detected. This can be useful in conjunction
            with duration=numpy.inf in e.g. Experiment.precondition to
            present instructions until the subject is ready to start.
        verbose=False -- Print to console as we go.
        '''
        super(Event,self).__init__()
        self.name = name
        self.duration = duration
        self.verbose = verbose
        self.setaslist('skiponresponse',skiponresponse)
        return

    def setaslist(self,name,val):
        '''
        Utility method for constraining an input val to list type before
        assigning as attribute name.
        '''
        try:
            setattr(self,name,list(val))
        except TypeError:
            setattr(self,name,[val])
        except:
            raise
        return

    def __call__(self,controller,endtime,currentevlog=None,currentresplog=None):
        '''
        Run the event through once and return pandas DataFrames with event
        and response logs.
        
        Keyword arguments:
        controller -- a Controller instance
        endtime -- the trial end in controller.clock units 
        currentevlog -- this variable is passed on to each callback function
            and can be used for e.g. response scoring. It can be
            constructed conveniently by concatenating eventlogs from
            previous trials (see EventSeq). We assume that each event is a
            single row in this log.
        currentresplog -- similar to evlog above, but one row per response
            rather than per event.

        Returns (both indexed by controller.clock()):
        eventlog -- pandas.DataFrame with 1 row. Columns contain the
            result of each callback (see event.eventkeys).
        resplog -- pandas.Series with one entry per key press.
        '''
        eventlog = prepeventrow([controller.clock()])
        eventlog['name'] = self.name
        eventlog['oncall'] = self.oncall(controller,currentevlog,
                currentresplog)
        if controller.eyetracker:
            controller.eyetracker.message(self.name)
        skipahead = False
        resplog = prepresprow()
        while controller.clock() < endtime and not skipahead:
            eventlog['onframe'] = self.onframe(controller,currentevlog,
                    currentresplog)
            response,resptime,frametime = controller();
            if len(response):
                thisresp = prepresprow(ind=resptime)
                thisresp['key'] = response
                thisresp['onresponse_score'],thisresp['onresponse_rt'] = \
                        self.onresponse(controller,response,resptime,
                                currentevlog,currentresplog)
                resplog = pandas.concat([resplog,thisresp],axis=0)
                if numpy.any(numpy.in1d(response,self.skiponresponse)):
                    skipahead = True
        eventlog['onend'] = self.onend(controller,currentevlog,currentresplog)
        if self.verbose:
            print eventlog.to_string(header=False)
        return eventlog,resplog

    def oncall(self,controller,currentevlog,currentresplog):
        '''This method is called at the beginning of the event.'''
        pass

    def onframe(self,controller,currentevlog,currentresplog):
        '''This method is called on the frame refresh.'''
        pass

    def onresponse(self,controller,response,rt,currentevlog,currentresplog):
        '''This method is called when a response is detected.'''
        nullresp = numpy.empty(numpy.shape(response))
        nullresp.fill(numpy.nan)
        return nullresp,nullresp

    def onend(self,controller,currentevlog,currentresplog):
        '''This method is called at the end of the event.'''
        pass

class EventSeq(Event):
    '''
    Base class for running through a fixed sequence of Event instances.
    Useful for defining conditions in complex designs, where your events
    might reflect stimulus, response, feedback and iti phases. In principle
    it is possible to run an entire experiment by specifying a long enough
    sequence in this instance, but in practice Experiment is a better bet
    for this. Direct use of this class is not supported - use the
    EventSeqRelTime and EventSeqAbsTime sub-classes.
    '''

    def __init__(self,events=[],**kwargs):
        '''
        Initialise a EventSeq instance.

        Keyword arguments:
        name -- is placed in the condition field in the res field generated
            on call.
        events -- list of Event instances. Note that these can be
            EventSeq instances to easily generate complex sequences
            (but note that if you use this functionality you want to set
            the name field to None in all but one recursion level, since
            all levels will write to the same condition field in the call
            return (and the outermost will be the last one in).
        duration -- duration of entire sequence in controller.clock units.
            If undefined the duration is inferred from the sum of the event
            durations. It can be useful to set the duration to a slightly
            larger value than this to correct for lag.
        verbose=False -- Print to console as we go.
        '''
        super(EventSeq,self).__init__(**kwargs)
        assert type(self) is not EventSeq, 'Direct use of super-class ' \
                'is not supported. Use derived sub-classes'
        self.events = events
        self.eventdur = numpy.array([x.duration for x in self.events])
        self.eventskip = [not len(x.skiponresponse) for x in self.events]
        return

    def __call__(self,controller=None,endtime=0.,currentev=None):
        '''
        do not use. This is for subclass use only.
        '''
        assert type(self) is not EventSeq, 'Direct use of super-class ' \
                'is not supported. Use derived sub-classes'
        if self.name:
            # the additional if statement here is useful to control when
            # existing condition fields get over-written (generally you
            # want that to happen at the condition level, but not at the
            # experiment level since you'd end up with a single label for
            # all events).
            currentev['condition'] = self.name
        # potential catchup phase
        controller.clock.waituntil(endtime)
        return currentev

class EventSeqRelTime(EventSeq):
    '''
    EventSeq subclass for relative timings. Useful for cases where absolute
    timing relative to experiment start is impossible (e.g., self timed
    tasks, synching to scanner pulses).
    '''
    def __init__(self,*args,**kwargs):
        super(EventSeqRelTime,self).__init__(*args,**kwargs)
        self.timing = self.eventdur
        assert not self.duration,\
                'cannot set EventSeq duration with relative timings'
        return

    def __call__(self,controller=None,endtime=0.,currentevlog=None,currentresplog=None):
        outevlog = prepeventrow()
        outresplog = prepresprow()
        for ind,ev in enumerate(self.events):
            # get time on every trial
            t = controller.clock()
            thisevlog,thisresp = ev(controller,t+self.timing[ind],
                    currentevlog,
                    currentresplog)
            # NB we append the full log, but do not return this to avoid
            # duplicates with nested calls
            currentevlog = pandas.concat([currentevlog,thisevlog],axis=0)
            currentresplog = pandas.concat([currentresplog,thisresp],
                    axis=0)
            outevlog = pandas.concat([outevlog,thisevlog],axis=0)
            outresplog = pandas.concat([outresplog,thisresp],axis=0)
            if self.verbose:
                print '%.1f\t %s' % (t,ev.name)
        return super(EventSeqRelTime,self).__call__(controller,endtime,outevlog),outresplog

class EventSeqAbsTime(EventSeq):
    '''
    EventSeq subclass for absolute timings. Given reasonable inputs this
    class will guarantee non-slip time over the course of the run.
    '''
    def __init__(self,*args,**kwargs):
        super(EventSeqAbsTime,self).__init__(*args,**kwargs)
        self.timing = numpy.cumsum(self.eventdur)
        realdur = numpy.sum(self.eventdur)
        if not self.duration:
            # infer duration from sum of eventdur
            # (note that by setting a duration > than the sum, you can
            # achieve a bit of catchup time to get around lag issues)
            self.duration = realdur
        assert numpy.isinf(self.duration) or self.duration >= realdur, \
                'event duration cannot exceed condition duration'
        assert not numpy.any(numpy.isinf(self.eventdur) | self.eventskip),\
                'EventSeqAbsTime is not supported for these events'
        return

    def __call__(self,controller,endtime=0.,currentevlog=None,currentresplog=None):
        starttime = controller.clock()
        evlog = prepeventrow()
        resplog = prepresprow()
        endtimes_trial = starttime + self.timing
        for ind,ev in enumerate(self.events):
            thisevlog,thisresp = ev(controller,endtimes_trial[ind],
                    currentevlog,currentresplog)
            evlog = pandas.concat([evlog,thisevlog],axis=0)
            resplog = pandas.concat([resplog,thisresp],axis=0)
            # NB we append the full log, but do not return this to avoid
            # duplicates with nested calls
            currentevlog = pandas.concat([currentevlog,thisevlog],axis=0)
            currentresplog = pandas.concat([currentresplog,thisresp],
                    axis=0)
        return super(EventSeqAbsTime,self).__call__(controller,endtime,evlog),resplog

class DrawEvent(Event):
    '''
    Event sub-class for handling a set of drawinstance, each of which have
    a draw() method that is called in turn on the frame rate. This is
    useful for visual presentation based on e.g. Psychopy.visual instances. 
    '''

    def __init__(self,drawinstances,**kwargs):
        '''
        Initialise a DrawEvent instance. The input drawinstances is a list
        of instances that have a draw method which takes no input
        arguments. Any remaining arguments are passed to Event.'''
        super(DrawEvent,self).__init__(**kwargs)
        self.setaslist('drawinstances',drawinstances)

    def onframe(self,controller,currentevlog,currentresplog):
        '''
        Callback for drawing the drawinstances to the screen. Usually
        referenced from the call method of Event.

        Arguments:
        controller -- a Controller instance
        currentevlog -- a result table (not actually used here but preserved
            for potential subclassing).
        currentresplog -- current responses.
        '''
        [x.draw() for x in self.drawinstances]
        return

class FeedbackEvent(DrawEvent):
    '''
    Provide feedback to the subject by plugging currentevlog and
    currentresplog into some scorer handle, and drawing different stimuli
    to the screen as a function of its return value.
    '''

    def __init__(self,drawinstances,scorer,correctdraw=[],incorrectdraw=[],omitdraw=[],**kwargs):
        super(FeedbackEvent,self).__init__(drawinstances,**kwargs)
        self.scorer = scorer
        # nb slice to ensure copy
        self.setaslist('commondraw',self.drawinstances[:])
        self.setaslist('correctdraw',correctdraw)
        self.setaslist('incorrectdraw',incorrectdraw)
        self.setaslist('omitdraw',omitdraw)
        return

    def oncall(self,controller,currentevlog,currentresplog):
        super(FeedbackEvent,self).oncall(controller,currentevlog,
                currentresplog)
        thisscore = self.scorer(currentevlog,currentresplog)
        # ensure copy
        self.drawinstances = self.commondraw[:]
        if pandas.isnull(thisscore):
            self.drawinstances += self.omitdraw
        elif thisscore==1:
            self.drawinstances += self.correctdraw
        else:
            self.drawinstances += self.incorrectdraw
        return thisscore

class DetectionEvent(DrawEvent):
    '''
    DrawEvent subclass for handling response scoring in simple detection
    tasks.
    '''

    def __init__(self,drawinstances,correct=None,minrt=0.,**kwargs):
        super(DetectionEvent,self).__init__(drawinstances,**kwargs)
        self.starttime = numpy.nan
        self.setaslist('correct',correct)
        assert not '*' in self.correct,'* is a reserved character'
        self.minrt = minrt
        return

    def oncall(self,controller,currentevent,currentresp):
        '''
        callback for DetectionEvent. The main functionality here is to
        reset the starttime property which can be used for calculating
        reaction times.

        Arguments:
        controller -- Controller instance
        currentevent -- not used here.
        currentresp -- not used here.
        '''
        # reset the timing at trial start
        super(DetectionEvent,self).oncall(controller,currentevent,
                currentresp)
        self.starttime = controller.clock()
        return

    def preparescore(self,response,resptime):
        rt = resptime-self.starttime
        # filter anticipations
        antind = rt < self.minrt
        rt[antind] = numpy.nan
        # can't use nan with char arrays
        response[antind] = '*'
        return response,rt

    def onresponse(self,controller,response,resptime,currentevlog,currentresplog):
        response,rt = self.preparescore(response,resptime)
        wascorrect = numpy.empty(numpy.shape(rt))
        wascorrect.fill(numpy.nan)
        wascorrect[numpy.in1d(response,self.correct)] = 1
        rt[numpy.isnan(wascorrect)] = numpy.nan
        return wascorrect,rt

class DecisionEvent(DetectionEvent):
    '''
    DetectionEvent sub-class for handling a combination of drawing stimuli to
    the screen (see DrawEvent) and collecting responses.'''

    def __init__(self,drawinstances,incorrect=None,**kwargs):
        '''
        Initialise a DecisionEvent instance and configure accuracy-based
        scoring.

        Arguments:
        drawinstances -- see DrawEvent
        correct=None -- list of keys to be scored as correct
        incorrect=None -- list of keys to be scored as incorrect
        minrt=0. -- minimum reaction time in controller.clock units for
            valid responses (quicker responses are ignored).
        kwargs -- any additional arguments are passed to Event.'''
        super(DecisionEvent,self).__init__(drawinstances,**kwargs)
        self.setaslist('incorrect',incorrect)
        assert not '*' in self.incorrect,'* is a reserved character'
        return

    def onresponse(self,controller,response,responsetime,currentevlog,currentresplog):
        '''
        callback for DecisionEvent. Provides basic accuracy-based scoring of
        responses.

        Arguments:
        controller -- Controller instance
        response -- list of inputs
        responsetime -- reaction times in absolute time units (controller.clock units)
        currentevlog -- not used here.
        currentresplog -- not used here.'''
        response,rt = self.preparescore(response,responsetime)
        wascorrect = numpy.empty(numpy.shape(rt))
        wascorrect.fill(numpy.nan)
        wascorrect[numpy.in1d(response,self.incorrect)] = 0
        wascorrect[numpy.in1d(response,self.correct)] = 1
        rt[wascorrect != 1] = numpy.nan
        if self.verbose:
            print 'correct=%s\tkey=%s' % (wascorrect,response)
        return wascorrect,rt

class NBackEvent(DecisionEvent):
    '''
    DecisionEvent subclass for scoring of Nback repetition detection tasks.
    '''
    def __init__(self,drawinstances,nback=1,nshift=0,**kwargs):
        super(NBackEvent,self).__init__(drawinstances,**kwargs)
        self.nback = nback
        # makes it possible to score e.g. an isi according to the previous
        # stimulus by shifting the whole scoring scheme a step backwards
        self.nshift = nshift
        self.wasrep = numpy.nan
        self.repkey = self.correct[:]
        self.notrepkey = self.incorrect[:]
        return

    def oncall(self,controller,currentevlog,currentresplog):
        # important to super here to reset starttime
        super(NBackEvent,self).oncall(controller,currentevlog,
                currentresplog)
        if self.nshift==0:
            currentname = self.name
        else:
            try:
                currentname = currentevlog.iloc[self.nshift]['name']
            except IndexError:
                currentname = numpy.nan
            except:
                raise
        try:
            previousname = currentevlog.iloc[self.nshift-self.nback]['name']
        except IndexError:
            previousname = numpy.nan
        except:
            raise
        self.wasrep = 0
        if pandas.isnull(currentname) or pandas.isnull(previousname):
            self.wasrep = numpy.nan
        elif currentname==previousname:
            self.wasrep = 1
        if self.verbose:
            print 'current=%s\t last=%s\twasrep=%s' % \
                    (currentname,previousname,self.wasrep)
        # so now we just reassign the keys
        if self.wasrep:
            self.correct = self.repkey[:]
            self.incorrect = self.notrepkey[:]
        else:
            self.correct = self.notrepkey[:]
            self.incorrect = self.repkey[:]
        return

class SynchEvent(DetectionEvent):
    '''
    DetectionEvent subclass for synchronising with the scanner. The typical
    use is to drop this instance in toward the end of a trial in a
    EventSeqRelTime sequence in order to trigger the next trial on the
    volume pulse.
    '''
    def __init__(self,drawinstances,targetkey,duration=numpy.inf,name='synch',**kwargs):
        super(SynchEvent,self).__init__(drawinstances,correct=targetkey,
                skiponresponse=targetkey,duration=duration,name=name,**kwargs)
        return

    def oncall(self,controller,currentevlog,currentresplog):
        super(SynchEvent,self).oncall(controller,currentevlog,
                currentresplog)
        if self.verbose:
            print 'waiting for pulse...'
        return

    def onresponse(self,controller,response,resptime,currentevlog,currentresplog):
        # nb we don't want to score RT of pulse for obvious reasons..
        response,null = self.preparescore(response,resptime)
        waspulse = numpy.empty(numpy.shape(null))
        waspulse.fill(numpy.nan)
        waspulse[numpy.in1d(response,self.correct)] = 1
        resptime[numpy.isnan(waspulse)] = numpy.nan
        return waspulse,resptime
